/***************************************************************************
 *   Copyright (C) 2008 by Robin Atwood   *
 *   robin@binro.org   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include "krcstat.h"
#include "constants.h"
#include "optionsdialogimp.h"

#include <qdrag.h>
#include <qmenu.h>
#include <qdir.h>
#include <qpainter.h>
#include <qlistview.h>
#include <qfile.h>
#include <qregexp.h>
#include <qsplitter.h>
#include <qtimer.h>
#include <qdebug.h>

#include <kglobal.h>
#include <klocale.h>
#include <kiconloader.h>
#include <kdeversion.h>
#include <kstatusbar.h>
#include <ktoolbar.h>
#include <ktabwidget.h>
#include <kconfig.h>
#include <kmessagebox.h>
#include <klineedit.h>
#include <kde_terminal_interface.h>
#include <klibloader.h>
#include <kpushbutton.h>
#include <kparts/part.h>


#include <kstdaccel.h>
#include <kaction.h>
#include <kstdaction.h>

#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>

using namespace std;

extern int errno;

KrcStat::KrcStat( QWidget* parent ) : KXmlGuiWindow( parent ),
      m_view( new KrcMainView(this) ),
      m_updtime( 10000 ),
			rlvlTB( "rlvlToolBar" )
{
	// Get the Gentoo base layout release level
	QFile statdir3( STATDIR3 );
	
	StatDir << "0" << STATDIR1;
	if ( statdir3.exists() )
		StatDir << STATDIR3;													// directory suddenly changed name
	else
		StatDir << STATDIR2;

	QString release; bool ok;
	QFile relfile( BASELVL );
  if ( relfile.open( QIODevice::ReadOnly ) ) 
	{
		QTextStream s( &relfile );
		QString line = s.readLine();
		release = line.section( ' ', 4, 4 );
		m_base = release.left( 1 ).toInt(&ok);
		relfile.close();
	}
	else
	{
		m_base = 1; release = "unknown";
	}
	
	setCaption( i18n("Runlevel Management"), false );

	// tell the KMainWindow that this is indeed the main widget
	setCentralWidget(m_view);
	
	// create the All view of services
	m_all = new KrcAllView( m_view->m_rlvlTab, m_base );

	// then, setup our actions
	setupActions();

	// and a status bar and some tool bars
	statusBar()->show();
	toolBar()->show();
	toolBar( rlvlTB )->show();

	// Apply the create the main window and ask the mainwindow to
	// automatically save settings if changed: window size, toolbar
	// position, icon size, etc.  Also to add actions for the statusbar
	// toolbar, and keybindings if necessary.

	setupGUI();
	setAutoSaveSettings();

	// get options
  m_config = KGlobal::config();
	m_colorGroup   = new KConfigGroup( m_config, "Colors" );
	readOptions();
	
	// get the current softlevel we booted with
	QFile file( StatDir[m_base] + "softlevel" );
  if ( file.open( QIODevice::ReadOnly ) ) 
	{
		QTextStream s( &file );
		m_softlevel = s.readLine();
		file.close();
	}
	else
	{
		int res = KMessageBox::warningContinueCancel( this, 
				i18n("No runlevel defined. Are you sure this is a Gentoo system?"));
		if ( res == KMessageBox::Cancel )
		{
			close();
		}
	}
	
	// set up status bar fields
	m_SBtext2 = new QLabel(  "softlevel: " + m_softlevel, this );
	m_SBtext2->setFrameShape( QFrame::StyledPanel );
	m_SBtext2->setFrameShadow( QFrame::Sunken );
	m_SBtext2->setToolTip( i18n("Runlevel in use. Set by the <b>softlevel</b> parameter at boot time") );
	statusBar()->addPermanentWidget(m_SBtext2);

	m_SBtext3 = new QLabel( "release: " + release, this );
	m_SBtext3->setFrameShape( QFrame::StyledPanel );
	m_SBtext3->setFrameShadow( QFrame::Sunken );
	m_SBtext3->setToolTip( i18n("The Gentoo Base System release") );
	statusBar()->addPermanentWidget(m_SBtext3);

	m_SBtext1 = new QLabel( i18n(""), this );
	m_SBtext1->setFrameShape( QFrame::StyledPanel );
	m_SBtext1->setFrameShadow( QFrame::Sunken );
	m_SBtext1->setToolTip( i18n("Service status summary") );
	statusBar()->addWidget(m_SBtext1);

	// set up icons
  KIconLoader il;
  m_px = new QPixmap( il.loadIcon( "krctat", KIconLoader::Desktop ) );
  m_minipx = new QPixmap( il.loadIcon( "krcstat", KIconLoader::Small ) );
  m_usedpx = new QPixmap( il.loadIcon( "service", KIconLoader::Small ) );
  m_unusedpx = new QPixmap( il.loadIcon( "service", KIconLoader::Small, 0, KIconLoader::DisabledState ) );
 
	// Set up watcher for service and runlevel status directories 
	m_FSW = new QFileSystemWatcher( this );
	m_FSW->addPath( StatDir[m_base] + "inactive" );
	m_FSW->addPath( StatDir[m_base] + "failed" );
	m_FSW->addPath( StatDir[m_base] + "starting" );
	m_FSW->addPath( StatDir[m_base] + "started" );
	m_FSW->addPath( StatDir[m_base] + "stopping" );
	m_FSW->addPath( SERVDIR );
	m_FSW->addPath( RLVLDIR );
	connect( m_FSW, SIGNAL(directoryChanged(const QString&)), this, SLOT(slotDirChanged(const QString&)) );
	connect( this, SIGNAL(needRefresh()), this, SLOT(slotRefresh()) );
	m_updating = false;
	
	// load the services into the left listbox
	m_firstView = true;
	loadServices( false );
	
	// Add runlevel child views
	loadRunLevels( false );
	
	// update status bar
	setStats();
	
	// indicate unused services
	updateServices();
	m_firstView = false;
        
	m_timer = new QTimer( this );
	connect( m_timer, SIGNAL(timeout()), this, SLOT(slotUpdate()) );
	m_timer->start( m_updtime );
	
	// open an embedded konsole to execute the commands
	m_view->m_term->startProgram( "/bin/sh", QStringList() << "-c" << "krcstat.sh" );
	connect( m_view->m_part, SIGNAL(destroyed()), this, SLOT(slotShellExit()) );
	
	QString info = i18n(
		"<h3>Available service list</h3>"
		"This is a list of the installed services. Simply drag "
    "the desired service to one of the runlevel panes and start it. <br><br>"
    "In the future the service will be started automatically at boot "
    "time (if that runlevel has been selected).");
		
	m_view->setWhatsThis( info );
}

KrcStat::~KrcStat()
{
	m_timer->stop();
}

void KrcStat::setupActions()
{
	m_actGrp = new QActionGroup( this );
	
	fileStart = newAction( "&Start", KIcon("media-playback-start"), Qt::CTRL+Qt::Key_S, this, "file_start" ); 
	fileStart->setToolTip( i18n("Start service") );	
	fileStart->setData( "start" );
	fileStart->setActionGroup( m_actGrp );
	
	fileStop = newAction( "Sto&p", KIcon("media-playback-stop"), Qt::CTRL+Qt::Key_P, this, "file_stop" );
	fileStop->setToolTip( i18n("Stop service") );	
	fileStop->setData( "stop" );
	fileStop->setActionGroup( m_actGrp );

	fileRestart = newAction( "&Restart", KIcon("media-skip-backward"), Qt::CTRL+Qt::Key_R, this, "file_restart");
	fileRestart->setToolTip( i18n("Restart service") );	
	fileRestart->setData( "restart" );
	fileRestart->setActionGroup( m_actGrp );

	if ( m_base == 2 )
	{
		filePauseResume = newAction( "Paus&e/Resume", KIcon("media-playback-pause"), Qt::CTRL+Qt::Key_E, this, "file_pause");
		filePauseResume->setToolTip( i18n("Stop/Start service ignoring dependencies") );	
	}
	else
	{
		filePauseResume = newAction( "Paus&e", KIcon("media-playback-pause"), Qt::CTRL+Qt::Key_E, this, "file_pause");
		filePauseResume->setToolTip( i18n("Stop service ignoring dependencies") );	
	}	

	filePauseResume->setData( "pause" );
	filePauseResume->setActionGroup( m_actGrp );

	fileZap = new KAction( KIcon("media-eject"), i18n("&Zap"), this );
	fileZap->setShortcut( Qt::CTRL+Qt::Key_Z );
	fileZap->setToolTip( i18n("Zap service") );	
	fileZap->setData( "zap" );
	fileZap->setActionGroup( m_actGrp );
	
	fileHelp = new KAction( KIcon("help-about"), i18n("&Help"), this );
	fileHelp->setShortcut( Qt::CTRL+Qt::Key_H );
	fileHelp->setToolTip( i18n("Zap service") );	
	fileHelp->setData( "help" );
	fileHelp->setActionGroup( m_actGrp );
	
	fileReload = new KAction( KIcon("edit-redo"), i18n("Re&load service configuration"), this );
	fileReload->setShortcut( Qt::CTRL+Qt::Key_L );
	fileReload->setToolTip( i18n("Reload service configuration") );	
	fileReload->setData( "reload" );
	fileReload->setActionGroup( m_actGrp );

 	connect( m_actGrp, SIGNAL(triggered(QAction*)), this, SLOT(slotCmdAct(QAction*)) ); 

	fileRefresh = newAction(i18n("Re&fresh status"), KIcon("view-refresh"), Qt::Key_F5, this, "file_refresh");
	connect(fileRefresh, SIGNAL(triggered(bool)), this, SLOT(slotRefresh())); 
	fileRefresh->setToolTip( i18n("Immediately refresh status") );	

	fileRemove = newAction( "Re&move", KIcon("edit-delete"), Qt::Key_Delete, this, "file_remove" );
	connect(fileRemove, SIGNAL(triggered(bool)), this, SLOT(slotRemove()) ); 
	fileRemove->setToolTip( i18n("Remove service from runlevel") );	

	KStandardAction::quit(kapp, SLOT(quit()), actionCollection());
	KStandardAction::preferences(this, SLOT(optionsPreferences()), actionCollection());
	KStandardAction::configureNotifications(this, SLOT(configureNotifications()), actionCollection());

	// set up a secondary toolbar for adding/removing runlevels

  rlvlAdd = newAction(i18n("&Add runlevel..."), KIcon("list-add"), Qt::CTRL+Qt::Key_A, this, "rlvl_add");
	connect(rlvlAdd, SIGNAL(triggered(bool)), this, SLOT(slotNewRlvl())); 
	rlvlAdd->setToolTip( i18n("Add new runlevel") );	
  
	m_rlvlEdit = new KLineEdit( this );
	m_rlvlEdit->setClearButtonShown( true );
	m_rlvlEdit->setClickMessage( i18n("New runlevel") );
	m_rlvlEdit->setMaximumWidth( 150 );
	m_rlvlEdit->setToolTip( i18n("Enter new runlevel name and press Enter") );

	rlvlEntry = new KAction( i18n(""), this );
 	rlvlEntry->setDefaultWidget( m_rlvlEdit );
	connect( m_rlvlEdit, SIGNAL(clearButtonClicked()), this, SLOT(slotEraseEdit()) ); 
	connect( m_rlvlEdit, SIGNAL(returnPressed(const QString&)), this, SLOT(slotEditRlvl(const QString&)) ); 

	rlvlDelete = newAction(i18n("R&emove runlevel"), KIcon("list-remove"), Qt::CTRL+Qt::Key_E, this, "rlvl_delete");
	connect(rlvlDelete, SIGNAL(triggered(bool)), this, SLOT(slotRmRlvl())); 
	rlvlDelete->setToolTip( i18n("Remove runlevel") );	

	m_actList1.append( (QAction *) rlvlDelete );
	m_actList2.append( (QAction *) rlvlEntry );
	m_actList2.append( (QAction *) rlvlDelete );
}

void KrcStat::saveOptions()
{
	m_colorGroup->writeEntry( "HSizes", m_view->sizes() );
	m_colorGroup->writeEntry( "VSizes", m_view->m_rightPane->sizes() );
	m_colorGroup->writeEntry( "inactive-color", textColors.inactive );
	m_colorGroup->writeEntry( "started-color", textColors.started );
	m_colorGroup->writeEntry( "starting-color", textColors.starting );
	m_colorGroup->writeEntry( "stopped-color", textColors.stopped );
	m_colorGroup->writeEntry( "stopping-color", textColors.stopping );
	m_colorGroup->writeEntry( "unused-color", textColors.unused );
	m_colorGroup->writeEntry( "failed-color", textColors.failed );
	m_colorGroup->writeEntry( "broken-color", textColors.broken );
	m_colorGroup->sync();
}

void KrcStat::readOptions()
{
  QList<int> hwidths = m_colorGroup->readEntry( "HSizes", QList<int>() );
  if ( !hwidths.empty() )
  	m_view->setSizes( hwidths );
  
  QList<int> vwidths = m_colorGroup->readEntry( "VSizes", QList<int>() );
  if ( !vwidths.empty() )
  	m_view->m_rightPane->setSizes( vwidths );
	else
		m_view->m_rightPane->setSizes( QList<int>() << 480 << 60 );
  
  textColors.inactive = m_colorGroup->readEntry( "inactive-color", QColor("gray").name() );
  textColors.started = m_colorGroup->readEntry( "started-color", QColor("green").name() );
  textColors.starting = m_colorGroup->readEntry( "starting-color", QColor("darkGreen").name() );
  textColors.stopped = m_colorGroup->readEntry( "stopped-color", QColor("red").name() );
  textColors.stopping = m_colorGroup->readEntry( "stopping-color", QColor("yellow").name() );
  textColors.unused = m_colorGroup->readEntry( "unused-color", QColor("darkCyan").name() );
  textColors.failed = m_colorGroup->readEntry( "failed-color", QColor("magenta").name() );
  textColors.broken = m_colorGroup->readEntry( "broken-color", QColor("cyan").name() );
}

KAction* KrcStat::newAction(const QString &text, const KIcon &icon, const QKeySequence &sc, QObject *parent, const QString &xml_text )
{
	KAction *action = new KAction( icon, text, parent ); 
	action->setShortcut( sc );
	actionCollection()->addAction( xml_text, action );
	return action;
}

bool KrcStat::queryExit()
{
  saveOptions();
  return true;
}

void KrcStat::slotRefresh()
{
	m_timer->stop();
	m_SBtext1->setText( i18n("Updating...") );
	kapp->processEvents();

	loadServices( true );
	loadRunLevels( true );
	updateServices();
	setStats();
	m_timer->start( m_updtime );
}

void KrcStat::slotUpdate()
{
	updateDaemons( "boot" );
	updateDaemons( "sysinit" );
	updateDaemons( m_softlevel );
	setStats();
}

void KrcStat::slotDirChanged( const QString &dir )
{
	m_updating = true;
	m_timer->stop();
	
	if ( dir == SERVDIR )
		loadServices( true );
	else
		loadRunLevels( true );

	updateServices();
	setStats();
	m_timer->start( m_updtime );
}

void KrcStat::slotCmdAct( QAction *action )
{
 	if ( m_saveName == "" ) return;
 	
 	QString arg = action->data().toString();
	if ( (arg == "pause") & (m_base == 2) )
	{
		if ( m_saveStatus == "started" ) 
			arg = "--nodeps stop";
		else
			arg = "--nodeps start";
	}
	else if ( arg == "help" ) arg = "--help";
	else if ( arg == "restart" )
	{
		if ( m_services.find( m_saveName )->m_status != "crashed" )
			arg = "--nodeps restart";
	}	
	
	QString cmd = SERVDIR + m_saveName + " " + arg + "\n";
	issueCommand( cmd );
}

void KrcStat::slotShellExit()
{
	m_view->m_part = 0;
	m_view->m_term = 0;
}

void KrcStat::slotNewRlvl()
{
	unplugActionList( "rlvl_actionlist" );
	plugActionList( "rlvl_actionlist", m_actList2 );
	rlvlAdd->setEnabled( false );
	m_rlvlEdit->setFocus();
}

void KrcStat::slotEditRlvl( const QString &newlvl )
{
	if ( newlvl != "" )
 	{
		QString newlvldir = RLVLDIR + newlvl,
		        cmd       = "mkdir -m 0755 " + newlvldir.toLatin1() + "\n";
    issueCommand( cmd );						
// 		if ( mkdir( newlvldir.toLatin1(), 0755 ) < 0 )
// 			KMessageBox::error( this, QString( i18n("Runlevel creation failed: ") ) + strerror( errno ) );
// 		else
// 		{
// 			//slotRefresh();
// 		}			
		m_rlvlEdit->clear();
	}

	unplugActionList( "rlvl_actionlist" );
	plugActionList( "rlvl_actionlist", m_actList1 );
	rlvlAdd->setEnabled( true );
}

void KrcStat::slotEraseEdit()
{
	m_rlvlEdit->clear();
}

void KrcStat::slotRemove()
{
	if ( m_saveRlvl == "unassigned" )
	{
		KMessageBox::error( this, i18n("You cannot remove an unassigned service!") );
		return;
	}
	
	if ( (QString( BOOTCRIT ).indexOf( " " + m_saveName + " " ) > -1) & (m_saveRlvl == "boot") )
	{
		KMessageBox::error( this, i18n("Your system may not function without this service!") );
		return;
	}
	
	issueCommand( "unlink " + QString( RLVLDIR ) + m_saveRlvl + "/" + m_saveName + "\n" );
// 	if ( unlink( (RLVLDIR + m_saveRlvl + "/" + m_saveName).toLatin1() ) < 0 )
// 		KMessageBox::error( this, QString( i18n("Service remove failed: ") ) + strerror( errno ) );
// 	else
// 	{	
		//slotRefresh();
	KrcService *serv = m_services.find( m_saveName );
	if ( serv > 0 )
		serv->m_rlvls.removeOne( m_saveRlvl );
// 	}	
}

void KrcStat::slotRmRlvl()
{
	m_curPage = m_view->m_rlvlTab->currentIndex();
	if ( (m_curPage+1) > (uint) m_runlevels.count() )
		return;                                  // all view
		
	QString rlvl = m_runlevels[m_curPage];                             
	
	if ( QString( RLVLCRIT ).indexOf( " " + rlvl + " " ) > -1 )
	{
		KMessageBox::error( this, i18n("Your system may not function without this runlevel!") );
		return;
	}
	
	int res = KMessageBox::warningContinueCancel( this, 
			i18n("Are you really sure you want to delete this runlevel (") + rlvl + ")?");
	if ( res == KMessageBox::Continue )
	{
		QDir rlvlDir( QString( RLVLDIR ) + rlvl, 0, 0, QDir::Files );
		QStringList assigned = rlvlDir.entryList();
		for ( QStringList::Iterator name = assigned.begin(); name != assigned.end(); ++name ) 
		{
			issueCommand( "unlink " + QString( RLVLDIR ) + rlvl + "/" + *name + "\n" );
// 			if ( unlink( (RLVLDIR + rlvl + "/" + (*name)).toLatin1() ) < 0 )
// 			{
// 				KMessageBox::error( this, QString( i18n("Runlevel remove failed: ") ) + strerror( errno ) );
// 				return;
// 			}	
// 			else
// 			{	
				KrcService *serv = m_services.find( *name );
				if ( serv > 0 )
					serv->m_rlvls.removeOne( rlvl );
// 			}	
		}

		issueCommand( QString( "rmdir " ) + RLVLDIR + rlvl + "\n" );
// 		if ( rmdir( (RLVLDIR + rlvl).toLatin1() ) < 0 )
// 			KMessageBox::error( this, QString( i18n("Runlevel remove failed: ") ) + strerror( errno ) );
// 		else
// 		{
		QWidget *w = m_view->m_rlvlTab->currentWidget();
		m_view->m_rlvlTab->removePage( w );
		delete w;
		m_runlevels.removeOne( rlvl );
		m_FSW->removePath( rlvlDir.absolutePath() );
// 		}
	}
}

void KrcStat::optionsPreferences()
{
	optionsDialogImp dlg( this, &textColors );
	if ( dlg.exec() == QDialog::Accepted )
	{
		saveOptions();
		slotRefresh();
	}
}

void KrcStat::configureNotifications() 
{
  KNotifyConfigWidget::configure( this );
}

void KrcStat::slotRMBItem( const QPoint &point )
{
	QTreeWidget *w = reinterpret_cast<QTreeWidget*>( m_view->m_rlvlTab->currentWidget() );
	QTreeWidgetItem *item = w->currentItem();
	if ( item == 0 ) return;

	m_saveName = item->text( 0 ).trimmed();
	m_saveStatus = item->text( 1 ).trimmed();
	m_saveRlvl = item->text( 3 ).trimmed();
	
	// check script for extra options
	
	m_popup = new QMenu( w );
	m_subpopup = 0;
	m_actSubGrp = 0;
	
	m_popup->addAction( fileStart );
	m_popup->addAction( fileStop );
	m_popup->addAction( fileRestart );
	m_popup->addAction( filePauseResume );
	m_popup->addAction( fileZap );
	m_popup->addAction( fileHelp );

	KrcService *serv = m_services.find(m_saveName );
	QString opts;
	
	if ( serv > 0 )	opts = serv->m_options;
	
	if ( !opts.isEmpty() )
	{
		if ( opts == "reload" )
		{
			m_popup->addAction( fileReload );
		}
		else if ( opts != "" ) 
		{
			QString menuitem, text;
			QAction *action;
	
			m_actSubGrp = new QActionGroup( this );
			connect( m_actSubGrp, SIGNAL(triggered(QAction*)), this, SLOT(slotCmdAct(QAction*)) ); 
			
			m_popup->addSeparator();
			m_subpopup = m_popup->addMenu( KIcon("system-run"), i18n("E&xtra options...") );
			
			for ( int p = 0;; ++p )
			{
				menuitem = opts.section( ' ', p, p );
				if ( menuitem == "" ) break;
				
				text = menuitem.left( 1 ).toUpper() + menuitem.mid( 1 );
				action = m_subpopup->addAction( KIcon("system-run"), text );
				action->setData( menuitem );
				action->setActionGroup( m_actSubGrp );
			}
		}		
	}

	m_popup->addSeparator();
	m_popup->addAction( fileRemove );
	m_popup->exec( w->viewport()->mapToGlobal(point) );    
	delete m_popup;
	if ( m_actSubGrp > 0 ) delete m_actSubGrp;
}

void KrcStat::slotRowChanged()
{
	QTreeWidget *w = reinterpret_cast<QTreeWidget*>( m_view->m_rlvlTab->currentWidget() );
	QTreeWidgetItem *item = w->currentItem();
	if ( item == 0 ) return;
	
	m_saveName = item->text( 0 ).trimmed();
	m_saveStatus = item->text( 1 ).trimmed();
	m_saveRlvl = item->text( 3 ).trimmed();
}

void KrcStat::loadServices( const bool reload  )
{
	QTreeWidget *sv = ( QTreeWidget* ) m_view->m_svcView ;
	if ( reload )
		sv->clear();
		
	KrcServiceList services; 
	QDir servDir( SERVDIR, 0, 0, QDir::Files );
	QStringList newserv = servDir.entryList();
	
	for ( QStringList::Iterator name = newserv.begin(); name != newserv.end(); ++name ) 
	{
		if ( ((*name).right( 3 ) != QString(".sh")) & ((*name).right( 1 ) != QString('~')) )
		{
			QTreeWidgetItem *item = new QTreeWidgetItem( sv );
			item->setText( 0, *name );
			item->setIcon( 0, *m_usedpx );
			
			KrcService newsvc( *name );
			KrcService *oldsvc = m_services.find( *name );
			
			if ( oldsvc != 0 )
			{
				newsvc.m_status = oldsvc->m_status;
				newsvc.m_laststatus = oldsvc->m_laststatus;
				newsvc.m_desc = oldsvc->m_desc;
				newsvc.m_daemon = oldsvc->m_daemon;
				if ( newsvc.m_daemon.isEmpty() )
					newsvc.m_daemon = getDaemon( *name );
				newsvc.m_options = oldsvc->m_options;
				newsvc.m_rlvls = oldsvc->m_rlvls;
				newsvc.m_cmd = oldsvc->m_cmd;
			}
			else
			{	
				newsvc.m_desc = getDescriptor( *name );
				newsvc.m_daemon = getDaemon( *name );
				newsvc.m_options = getOptions( *name );
			}	
			
			item->setToolTip( 0, servInfo( newsvc ) );
			services.append( newsvc );
		}
  }
  
  m_services = services;
}

void KrcStat::loadRunLevels( const bool reload )
{
	QDir inactiveDir( StatDir[m_base] + "inactive", 0, 0, QDir::Files );
	QDir startedDir(  StatDir[m_base] + "started",  0, 0, QDir::Files );
	QDir startingDir( StatDir[m_base] + "starting", 0, 0, QDir::Files );
	QDir stoppingDir( StatDir[m_base] + "stopping", 0, 0, QDir::Files );
	
	runStatus.inactive =  inactiveDir.entryList();
	runStatus.started  =  startedDir.entryList();
	runStatus.starting =  startingDir.entryList();
	runStatus.stopping =  stoppingDir.entryList();
	
	QDir rlvlsDir( RLVLDIR, 0, 0, QDir::Dirs|QDir::NoDotAndDotDot);
	m_runlevels = rlvlsDir.entryList();
	
	int page = 0; uint i;
	
	if ( reload )
		m_curPage = m_view->m_rlvlTab->currentIndex();
	
	m_all->clear();
	for ( int k = 0; k < m_services.count(); ++k )
		m_services[k].m_rlvls.clear();
		
	for ( i = 0; i < (uint) m_runlevels.count(); ++i ) 
	{
		m_rlvl = (KrcRlvlView *) m_view->m_rlvlTab->widget( i );
		if ( (m_rlvl == 0) | (i+1 == (uint) m_view->m_rlvlTab->count()) )
		{																									// allocate a new page and listview
			m_rlvl = new KrcRlvlView( m_view->m_rlvlTab, this );
			m_view->m_rlvlTab->insertTab( i, m_rlvl, m_runlevels[i] );
			connect( m_rlvl, SIGNAL(customContextMenuRequested(const QPoint &)),
							this, SLOT(slotRMBItem(QPoint)) );
			connect( m_rlvl, SIGNAL(itemSelectionChanged()), this, SLOT(slotRowChanged()) );
		}
		else				
		{																									// reuse an old one
			m_view->m_rlvlTab->setTabText( i, m_runlevels[i] );
			m_rlvl->clear();
		}
		
		m_rlvl->load( m_runlevels[i] );
		resizeColumns( m_rlvl );
		m_rlvl->hideColumn( 3 );                   // not really helpful here
		
		QString abspath( RLVLDIR + m_runlevels[i] );
		if ( m_FSW->directories().indexOf( abspath ) == -1 )
			m_FSW->addPath( abspath ); 							 // watch for new services
		
		if ( (m_runlevels[i] == m_softlevel) & m_firstView ) m_curPage = page;
		++page;
	}
	
	// remove unneeded tabs if runlevels removed
	while ( m_view->m_rlvlTab->count() > m_runlevels.count()+1 )
	{
		QWidget *w = m_view->m_rlvlTab->widget( i );
		m_view->m_rlvlTab->removePage( w );
		delete w;
		++i;
	}
	
	// add unassigned services to the ALL view
	for ( uint i = 0; i < (uint) m_services.count(); ++i ) 
	{
		if ( m_services[i].m_rlvls.count() == 0  )
		{
			QString status, daemon, desc;
			getStatus( m_services[i].m_name, status, daemon, desc );
			if ( status == "stopped" )
				status = "unused";
			new KrcListItem( m_all, m_services[i].m_name, status, daemon, "unassigned", desc, &textColors );
		}	
	}
	
	if ( !reload )
	{
		m_view->m_rlvlTab->addTab( m_all, "all" );

		connect( m_all, SIGNAL(customContextMenuRequested(const QPoint &)),
						this, SLOT(slotRMBItem(QPoint)) );
		connect( m_all, SIGNAL(itemSelectionChanged()), this, SLOT(slotRowChanged()) );
	}
	
	resizeColumns( m_all );
	m_all->sortItems( 0, Qt::AscendingOrder );
	m_view->m_rlvlTab->setCurrentIndex( m_curPage );
}

void KrcStat::resizeColumns( QTreeWidget *tw )
{
	for ( int k = 0; k < tw->columnCount(); ++k )
		tw->resizeColumnToContents( k );
}

void KrcStat::updateServices()
{
	QTreeWidget *tw = reinterpret_cast<QTreeWidget*>( m_view->m_svcView );
	QTreeWidgetItem *item;
	KrcService *serv;	
	
	for ( int k = 0; k < tw->topLevelItemCount(); ++k ) 
	{
		item = tw->topLevelItem( k );
		QString name = item->text( 0 ).trimmed();
		serv = m_services.find( name );
		if ( serv == NULL ) continue;
		
		if ( serv->m_rlvls.count() == 0 )
		{
			item->setIcon( 0, *m_unusedpx );
			if ( serv->m_status == "stopped" )
			{	
				serv->m_status = "unused";
				serv->m_laststatus = "unused";
			}
		}
		else
		{
			item->setIcon( 0, *m_usedpx );
		}	
		
		if ( !m_firstView & (serv->m_status != serv->m_laststatus) )
		{
			KNotification::event( "service_" + serv->m_status, 
					serv->m_name + i18n(" has changed status to ") + serv->m_status,	QPixmap(), m_view );
		}	
	}
}

void KrcStat::updateDaemons( const QString &rlvl )
{
	const int p = m_runlevels.indexOf( rlvl );
	QTreeWidget *view = reinterpret_cast<QTreeWidget*>( m_view->m_rlvlTab->widget( p ) );
	QTreeWidgetItem *item;
	bool changed = false;
	
	for ( int k = 0; k < view->topLevelItemCount(); ++k ) 
	{
		item = view->topLevelItem( k );
		QString name   = item->text( 0 ).trimmed(),
						status = item->text( 1 ).trimmed(),
						daemon = item->text( 2 ).trimmed();

		if ( (daemon != "") & (status == "started") )
		{
			KrcService *serv = m_services.find( name );
			if ( serv == NULL ) continue;
			
			if ( !pidExists( serv->m_cmd ) )
			{
				serv->m_status = "crashed";
				serv->m_laststatus = "crashed";
				
				m_note = KNotification::event( "service_" + serv->m_status, 
						serv->m_name + i18n(" has changed status to ") + serv->m_status,
						QPixmap(), m_view );
				//connect( m_note, SIGNAL(closed()), this , SLOT(slotFailedNote()) );		

				changed = true;
			}
		}
	}
	
	if ( changed )
		loadRunLevels();
}

void KrcStat::slotFailedNote()
{
	qDebug() << "slotFailedNote entered" << endl;
}

void KrcStat::getStatus( const QString &svc, QString &status, QString &daemon, QString &desc )
{
	KrcService *serv = m_services.find( svc );
	int j;
	
	if ( serv == 0 ) return;
	
	if ( runStatus.inactive.indexOf( svc ) > -1 )
		status = "inactive";
	else if ( runStatus.started.indexOf( svc ) > -1 )
		status = "started";
	else if ( runStatus.starting.indexOf( svc ) > -1 )
		status = "starting";
	else if ( runStatus.stopping.indexOf( svc ) > -1 )
		status = "stopping";
	else
		status = "stopped";
			
	if ( serv->m_daemon.isEmpty() )
		serv->m_daemon = getDaemon( serv->m_name );
	daemon = serv->m_daemon;

	if ( (status == "started") & (daemon != "") & (daemon != "*NONE*") )
	{
		if ( !pidExists( daemon ) )
		{
			status = "crashed";
		}
	}
	else if ( (status == "started") & (daemon == "") )
		serv->m_daemon = "*NONE*";
	
	if ( daemon != "" )
	{
		serv->m_cmd = daemon;
		j = daemon.length() - daemon.lastIndexOf( "/" ) - 1;
		daemon = daemon.right( j );
	}

	desc = serv->m_desc;
	
	if ( serv > 0 )
	{
		if ( serv->m_laststatus == "" )
			serv->m_laststatus = status;
		else
			serv->m_laststatus = serv->m_status;
		serv->m_status = status;
	}
}

void KrcStat::setStats()
{
	int started = 0, stopped = 0, unused = 0, crashed = 0;
	QString status, rlvl;
	
	for ( int k = 0; k < m_all->topLevelItemCount(); ++k )
	{
		QTreeWidgetItem *item = m_all->topLevelItem( k );
		status = item->text( 1 );
		rlvl = item->text( 3 );
		if ( (rlvl == "boot") | (rlvl == m_softlevel) | (rlvl == "unassigned") |
				 (rlvl == "shutdown") | (rlvl == "sysinit") )
		{
			if ( status == "started" )
				started += 1;
			else if ( status == "stopped" )
				stopped += 1;
			else if ( status == "unused" )
				unused += 1;
			else if ( status == "crashed" )
				crashed += 1;
		}
	}
	m_SBtext1->setText( 
		QString(i18n("services: %1 started, %2 stopped, %3 crashed, %4 unused"))
				.arg(started).arg(stopped).arg(crashed).arg(unused) );
}

QString KrcStat::getDescriptor( const QString &svc )
{
	QString desc;
	int j, k;
	
	if ( m_base == 2 )
	{
		QStringList output; QString line;
		output = sysCommand( QString("grep description= ") + SERVDIR + svc );
		if ( !output.isEmpty() )
		{
			line = output[0].trimmed();
			j = line.indexOf( '=' ) + 2; 
			k = line.length() - 1;
			desc = line.mid( j, k-j );
		}
		else
			desc = "";
	}
	else
		desc = "*NONE*";
	
	return desc;
}

QString KrcStat::getDaemon( const QString &svc )
{
	QString dmnfile, daemon; 
	int j, k;
	
	if ( m_base == 1 )
		dmnfile = StatDir[m_base] + "daemons/" + svc;
	else
		dmnfile = StatDir[m_base] + "daemons/" + svc + "/001";

	QFile file( dmnfile );
  if ( file.open( QIODevice::ReadOnly ) ) 
	{
		QTextStream s( &file );
		QString l = s.readLine();
		file.close();
		if ( m_base == 1 )
		{
			j = l.indexOf( "'" ) + 1; 
			k = l.indexOf( "'", j );
			daemon = l.mid( j, k-j );
		}
		else
		{
			j = l.indexOf('=') + 1;
			k = l.length() - j;
			daemon = l.right( k );
		}
	}
	else
		daemon = "";

	return daemon;
}

QString KrcStat::getOptions( const QString &svc )
{
	const QString script = SERVDIR + svc;
	QString opts = extractCommands( "opts", script );
	
	if ( opts.isEmpty() )
	{
		opts = extractCommands( "extra_commands", script );
		opts += " " + extractCommands( "extra_started_commands", script );
		opts += " " + extractCommands( "extra_stopped_commands", script );
	}
	
	return opts.simplified();
}

QString KrcStat::extractCommands( const QString &variable, const QString &script)
{
	QString cmd, opts = "";
	QStringList output; 
	const int l = variable.length();
	
	cmd = "grep ^\\ *" + variable +"= " + script;
	output = sysCommand( cmd );

	if ( !output.isEmpty() )
	{
		if ( output[0].startsWith( variable + "=\"" ) )
		{
			opts = output[0].mid( l+2, output[0].length()-l-3 );
			const int p = opts.indexOf( "${" + variable + "}" );
			if ( p > -1 )
				opts = opts.remove( p, l+3 );
		}
	}

	return opts;
}

QString KrcStat::servInfo( const KrcService &service )
{
	QString info;
	
	info = "<h4>Service Information</h4>";
	info += "<dl compact><dt><b>File:</b><dd>" + QString(SERVDIR) + service.m_name;
	if ( !service.m_options.isEmpty() )
		info += "<dt><b>Options:</b><dd>" +	service.m_options;
	if ( !service.m_desc.isEmpty() )
		info += "<dt><b>Description:</b><dd>" + service.m_desc;
	info += "</dl>";
	
	return info;
}


QStringList KrcStat::sysCommand( const QString &cmd )
{
	FILE *fp; QStringList out; 
	char line[1024]; QChar eol( '\n' );
	 	
	fp = popen( cmd.toLatin1(), "r" );
	while ( fgets( line, sizeof( line ), fp ) )
	{
		QString s = line;
		if ( s.endsWith( QString( eol ) ) )
			s.truncate( s.length() - 1 );	
		out << s;
	}
	pclose(fp);
	return out;
}

void KrcStat::issueCommand( const QString &cmd )
{
	if ( m_view->m_term > 0 )
		m_view->m_term->sendInput( cmd );
	else
		sysCommand( "sudo " + cmd );
}

bool KrcStat::pidExists( const QString &daemon )
{
	bool exists = false;
	QStringList pid = sysCommand( "pidof -s " + daemon );
	
	if ( !pid.isEmpty() )
	{ 																					// double check in case it's a script
		if ( pid[0].isNull() )
		{																					// blank line returned
			pid = sysCommand( "ps -ef | grep " + daemon + " | grep -v grep" );
			exists = !pid.isEmpty();
		}
		else
			exists = true;
	}
	else																				// no output returned
	{
		pid = sysCommand( "ps -ef | grep " + daemon + " | grep -v grep" );
		exists = !pid.isEmpty();
	}
	
	return exists;
}

/*void KrcStat::execKonsole( const QString &cmd )
{
	KProcess *proc = new KProcess( this );
	*proc << "konsole";
	*proc << "--noclose" << "-e" << cmd;
	qDebug() << cmd << endl;
	
	proc->execute();
	delete proc;
}*/

#include "krcstat.moc"
